#include <CL/sycl.hpp>

#include <algorithm>
#include <iostream>
#include <vector>

/*
* This program reproduces a bug found on my system, where
* for _some_ image sizes, reading the _last_ pixel causes a deadlock.
*/
std::vector<int> affectedImageSizes = { {} };
/*
* I used this procedure to detect the image sizes affected by this bug:
* - run the program
* - if the program deadlocks (the last log in the console is: 'writeback ...'):
*     add the corresponding image size to 'affectedImageSizes' and repeat.
*
* On my system the result is: affectedImageSizes = { {256, 512, 768, 1024, 1280, 1536, 1792} };
* i.e every image whose size is a multiple of 256 is affected by the bug.
*/

constexpr uint32_t Red = 0b10101100111100001111011100001000;
constexpr uint32_t Green = 0b10111101101101001101011000111000;
constexpr uint32_t Blue = 0b00101110111100011101011110001001;
constexpr uint32_t Alpha = 0b00101000111100001101011101001000;

int main()
{
  using namespace cl::sycl;

  auto ehandler = [](exception_list exceptionList) {
    for (std::exception_ptr const& e : exceptionList)
    {
      try
      {
        std::rethrow_exception(e);
      }
      catch (exception const& e)
      {
        std::terminate();
      }
    }
  };

  queue q(gpu_selector(), ehandler);

  int const nMaxItemsPerWorkgroup =
    q.get_device().get_info<info::device::max_work_group_size>();

  for (int nPixels = 1; nPixels < 2000; ++nPixels)
  {
    std::cout << "[With " << nPixels << " pixels]" << std::endl;
    bool const readLastPixel = std::find(affectedImageSizes.begin(), affectedImageSizes.end(), nPixels) == affectedImageSizes.end();

    // Build the input data

    std::vector<uint32_t> rgba;
    rgba.reserve(nPixels * 4);

    for (int i = 0; i < nPixels; ++i)
    {
      rgba.push_back(Red);
      rgba.push_back(Green);
      rgba.push_back(Blue);
      rgba.push_back(Alpha);
    }

    image<1> img(
      rgba.data(),
      image_channel_order::rgba,
      image_channel_type::unsigned_int32,
      nPixels);

    // Verify that the image has the same size (in bytes) as the vector
    if (img.get_size() != sizeof(uint32_t) * rgba.size())
      throw std::logic_error("size mismatch");

    // Build the output data

    std::vector<uint32_t> rgbaCopy(rgba.size(), 0);

    // Use a kernel to copy the input data to the output data
    {
      buffer<uint32_t, 1> imgDataOutBuffer(
        rgbaCopy.data(),
        rgbaCopy.size());

      int const nMaxPixelsPerThread = 1 + (nPixels - 1) / nMaxItemsPerWorkgroup;
      int const pixelStride = 1 + (nPixels - 1) / nMaxPixelsPerThread;

      std::cout << "call kernel with nPixels = '" << nPixels << "', pixelStride '" << pixelStride << "' ..." << std::endl;

      q.submit([&](handler& h) {
        auto aImg = img.get_access<cl::sycl::cl_int4, access::mode::read>(h);
        auto aOutput = imgDataOutBuffer.get_access<access::mode::write>(h);
        auto const r = nd_range<1>{
          range(pixelStride),
          range(pixelStride)
        };
        h.parallel_for<class ReproMinimalDeadlock_Kernel>(
          r,
          [=](nd_item<1> it) {
            // do not read the last pixel if the image size is affected by the bug
            int const endPixel = nPixels - (readLastPixel ? 0 : 1);
            for (int i = it.get_local_id(); i < endPixel; i += pixelStride)
            {
              auto pixel = aImg.read(i);
              aOutput[4 * i + 0] = pixel.s0();
              aOutput[4 * i + 1] = pixel.s1();
              aOutput[4 * i + 2] = pixel.s2();
              aOutput[4 * i + 3] = pixel.s3();
            }
          });
        });
      q.wait();
      std::cout << "writeback ..." << std::endl;
    }
    std::cout << "writeback done." << std::endl;

    // Verify that the ouptut data matches the input data.
    // Skip the last pixel if the image size is affected by the bug.
      
    for (int i = 0, sz = rgbaCopy.size() - (readLastPixel ? 0 : 4); i < sz; ++i)
      if (rgba[i] != rgbaCopy[i])
        throw std::logic_error("in != out");
  }

  return 0;
}